/* -*-Mode:C; tab-width:4; indent-tabs-mode:t; c-file-style:"stroustrup";-*- */
// $Id: task.c,v 1.34 2003/11/06 14:36:33 rick_price Exp $

#include <PalmOS.h>
#include "task.h"
#include "progect.h"
#include "xb.h"
#include "progectdb.h"
#include "progectRsc.h"
#include "ToDoDB.h"
#include "MemoDB.h"
#include "AddressDB.h"

void dbgPrintTask(DmOpenRef dbP, UInt16 index)
{
	MemHandle h;
	TaskRecordType *p;

	DBGMSG((DBB, "dbgPrintTask"));

	h = DmQueryRecord(dbP, index);
	p = MemHandleLock(h);
	DBGMSG((DBB, "Complete record next line"));
	DBGMSGBIN((DBB, p, MemHandleSize(h)));
	MemHandleUnlock(h);
}



/****************************************************************************
 * Name : TaskSetAttr
 * Desc : set the attributes of a task
 * Parm : 
 * 			-> db to search
 * 			-> index of record
 * 			-> new attributes
 * Out  : pgErr
 * Auth : lb, 22.08.2000
 * Mod  : lb, 2001-01-21
 * 			changing attr should not set dirty (move or change level)
 ***************************************************************************/
pgErr TaskSetAttr(DmOpenRef dbP, UInt16 index, TaskAttrType attr)
{
	MemHandle h;
	TaskRecordType *p;

	DBGMSG((DBB, "TaskSetAttr"));

	h = DmGetRecord(dbP, index);

	p = MemHandleLock(h);
	DmWrite(p, OffsetOf(TaskRecordType, attr), &attr, sizeof(UInt16));
	MemHandleUnlock(h);
	DmReleaseRecord(dbP, index, false);

	return pgOK;
} // pgErr TaskSetAttr(DmOpenRef dbP, UInt16 index, TaskAttrType attr)



/****************************************************************************
 * Name : TaskGetAttr
 * Desc : get the attributes of a task
 * Parm : 
 * 			-> db to search
 * 			-> index of record
 * 			-> pointer to TaskAttrType struct to assign values to
 * Out  : attributes,Boolean result - depending on success or failure of func
 * Auth : lb, 22.08.2000
 ***************************************************************************/
Boolean TaskGetAttr(DmOpenRef dbP, UInt16 index,TaskAttrType *attr)
{
	MemHandle h;
	TaskRecordType *p;

	DBGMSG((DBB, "TaskGetAttr"));

	h = DmQueryRecord(dbP, index);
	if (!h)
		return false;

	p = MemHandleLock(h);
	*attr = p->attr.bits;
	MemHandleUnlock(h);

	return true;
} // Boolean TaskGetAttr(DmOpenRef dbP, UInt16 index,TaskAttrType *attr)



/****************************************************************************
 * Name : TaskSetFormat
 * Desc : set the format of a task
 * Parm : 
 * 			-> db to search
 * 			-> index of record
 * 			-> new format
 * Out  : pgErr
 * Auth : lb, 22.08.2000
 ***************************************************************************/
pgErr TaskSetFormat(DmOpenRef dbP, UInt16 index, TaskFormatType format)
{
	MemHandle h;
	TaskRecordType *p;

	DBGMSG((DBB, "TaskSetFormat"));

	h = DmGetRecord(dbP, index);

	p = MemHandleLock(h);
	DmWrite(p, OffsetOf(TaskRecordType, format), &format, sizeof(UInt16));
	MemHandleUnlock(h);
	DmReleaseRecord(dbP, index, true);

	return pgOK;
} // pgErr TaskSetFormat(DmOpenRef dbP, UInt16 index, TaskFormatType format)



/****************************************************************************
 * Name : TaskGetFormat
 * Desc : get the format of a task
 * Parm : 
 * 			-> db to search
 * 			-> index of record
 * Out  : format
 * Auth : lb, 26.08.2000
 ***************************************************************************/
TaskFormatType TaskGetFormat(DmOpenRef dbP, UInt16 index)
{
	MemHandle h;
	TaskRecordType *p;
	TaskFormatType format;

	DBGMSG((DBB, "TaskGetFormat"));

	h = DmQueryRecord(dbP, index);

	p = MemHandleLock(h);
	format = p->format.bits;
	MemHandleUnlock(h);

	return format;
} // TaskFormatType TaskGetFormat(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSetLevel
 * Desc : set the level of a task
 * Parm : 
 * 			-> db to search
 * 			-> index of record
 * 			-> new level (0-31)
 * Out  : pgErr
 * Auth : lb, 28.07.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 ***************************************************************************/
pgErr TaskSetLevel(DmOpenRef dbP, UInt16 index, UInt8 level)
{
	TaskAttrType attr;

	DBGMSG((DBB, "TaskSetLevel"));

	if (level > 31)
		return pgError;

	if(!TaskGetAttr(dbP, index, &attr))
		return pgError;

	attr.level = level;
	TaskSetAttr(dbP, index, attr);

	return pgOK;
} // pgErr TaskSetLevel(DmOpenRef dbP, UInt16 index, UInt8 level)



/****************************************************************************
 * Name : TaskGetLevel
 * Desc : get the level of a task
 * Parm : 
 * 			-> db to search
 * 			-> index of record
 * Out  : level, 0 if not found
 * Auth : lb, 27.07.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 ***************************************************************************/
UInt8 TaskGetLevel(DmOpenRef dbP, UInt16 index)
{
	TaskAttrType attr;

	DBGMSG((DBB, "TaskGetLevel"));

	if(!TaskGetAttr(dbP, index, &attr))
		return 0;
	else
		return attr.level;
} // UInt8 TaskGetLevel(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSetPriority
 * Desc : set the priority of a task
 * Parm : 
 * 			-> db to search
 * 			-> index of record
 * 			-> new priority
 * Out  : 
 * Auth : lb, 06.08.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 * Mod	: lb, 2001-09-10
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
pgErr TaskSetPriority(DmOpenRef dbP, UInt16 index, UInt8 priority)
{
	MemHandle h;
	TaskExtendedRecordType *p;

	DBGMSG((DBB, "TaksSetPriority"));

	h = DmGetRecord(dbP, index);
	p = MemHandleLock(h);
	DmWrite(p,
		StdFieldsOffset(p) + OffsetOf(TaskStandardFields, priority), 
		&priority, sizeof(UInt8));
	MemHandleUnlock(h);
	DmReleaseRecord(dbP, index, true);
	return pgOK;
} // void TaskSetPriority(DmOpenRef dbP, UInt16 index, UInt8 priority)



/****************************************************************************
 * Name : TaskGetPriority
 * Desc : get the priority of a task
 * Parm : 
 * 			-> db to search
 * 			-> index of record
 * Out  : level, 0 if not found
 * Auth : lb, 06.08.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 * Mod	: lb, 2001-09-10
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
UInt8 TaskGetPriority(DmOpenRef dbP, UInt16 index)
{
	MemHandle h;
	TaskExtendedRecordType *p;
	UInt8 priority;

	DBGMSG((DBB, "TaskGetPriority"));

	h = DmQueryRecord(dbP, index);
	p = MemHandleLock(h);
	priority = StdFields(p)->priority;
	MemHandleUnlock(h);
	if (priority == 0)
		priority = NO_PRIORITY;
	return priority;
} // UInt8 TaskGetPriority(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSetCategory
 * Desc : set the category of a task
 * Parm : 
 * 			-> db to search
 * 			-> index of record
 * 			-> new category
 * Out  : 
 * Auth : rw, 30.12.2000
 ***************************************************************************/
pgErr TaskSetCategory(DmOpenRef dbP, UInt16 index, UInt16 category)
{
	UInt16 attr;

	DBGMSG((DBB, "TaskSetPriority"));

	DmRecordInfo(dbP, index, &attr, NULL, NULL);
	attr = (attr & ~dmRecAttrCategoryMask) | category | dmRecAttrDirty;
	DmSetRecordInfo(dbP, index, &attr, NULL);

	return pgOK;
} // void TaskSetCategory(DmOpenRef dbP, UInt16 index, UInt16 category)



/****************************************************************************
 * Name : TaskGetCategory
 * Desc : get the category of a task
 * Parm : 
 * 			-> db to search
 * 			-> index of record
 * Out  : category, 0 if Unfiled
 * Auth : rw, 30.12.2000
 ***************************************************************************/
UInt16 TaskGetCategory(DmOpenRef dbP, UInt16 index)
{
	UInt16 attr;
	UInt16 category;

	DBGMSG((DBB, "TaskGetCategory"));

	DmRecordInfo(dbP, index, &attr, NULL, NULL);
	category = attr & dmRecAttrCategoryMask;
	return category;
} // UInt16 TaskGetCategory(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSetHasChild
 * Desc : set the hasChild field of task
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * 			-> value of hasChild field
 * Out  : 
 * Auth : lb, 28.07.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 ***************************************************************************/
pgErr TaskSetHasChild(DmOpenRef dbP, UInt16 index, Boolean hasChild)
{
	TaskAttrType attr;

	DBGMSG((DBB, "TaskSetHasChild"));

	if(!TaskGetAttr(dbP, index, &attr))
		return pgError;
	
	attr.hasChild = hasChild;
	TaskSetAttr(dbP, index, attr);

	return pgOK;
} // pgErr TaskSetHasChild(DmOpenRef dbP, UInt16 index, Boolean hasChild)



/****************************************************************************
 * Name : TaskGetHasChild
 * Desc : get the hasChild field of task
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * Out  : hasChild field value
 * Auth : lb, 28.07.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 ***************************************************************************/
Boolean TaskGetHasChild(DmOpenRef dbP, UInt16 index)
{
	TaskAttrType attr;

	DBGMSG((DBB, "TaskGetHasChild"));

	if(!TaskGetAttr(dbP, index, &attr))
		return 0;
	else
		return attr.hasChild;
} // Boolean TaskGetHasChild(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSetHasPrev
 * Desc : set the firstChild field of task
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * 			-> value of hasPrev field
 * Out  : 
 * Auth : lb, 28.07.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 ***************************************************************************/
pgErr TaskSetHasPrev(DmOpenRef dbP, UInt16 index, Boolean hasPrev)
{
	TaskAttrType attr;

	DBGMSG((DBB, "TaskSetHasPrev"));

	if(!TaskGetAttr(dbP, index, &attr))
		return pgError;
	attr.hasPrev = hasPrev;
	TaskSetAttr(dbP, index, attr);

	return pgOK;
} // pgErr TaskSetHasPrev(DmOpenRef dbP, UInt16 index, Boolean hasPrev)



/****************************************************************************
 * Name : TaskGetHasPrev
 * Desc : get the firstChild field of task
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * Out  : firstChild field value
 * Auth : lb, 28.07.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 ***************************************************************************/
Boolean TaskGetHasPrev(DmOpenRef dbP, UInt16 index)
{
	TaskAttrType attr;

	DBGMSG((DBB, "TaskGetHasPrev"));

	if(!TaskGetAttr(dbP, index, &attr))
		return 0;
	else
		return attr.hasPrev;
} // Boolean TaskGetHasPrev(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskGetHasNext
 * Desc : get the hasNext field of task
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * Out  : hasNext field value
 * Auth : lb, 28.07.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 ***************************************************************************/
Boolean TaskGetHasNext(DmOpenRef dbP, UInt16 index)
{
	TaskAttrType attr;

	DBGMSG((DBB, "TaskGetHasNext"));

	if(!TaskGetAttr(dbP, index, &attr))
		return 0;
	else
		return attr.hasNext;
} // Boolean TaskGetHasNext(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSetHasNext
 * Desc : set the hasNext field of task
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * 			-> next field
 * Out  : 
 * Auth : lb, 28.07.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 ***************************************************************************/
pgErr TaskSetHasNext(DmOpenRef dbP, UInt16 index, Boolean next)
{
	TaskAttrType attr;

	DBGMSG((DBB, "TaskSetHasNext"));

	if(!TaskGetAttr(dbP, index, &attr))
		return pgError;
	attr.hasNext = next;
	TaskSetAttr(dbP, index, attr);

	return pgOK;
} // pgErr TaskSetHasNext(DmOpenRef dbP, UInt16 index, Boolean next)



/****************************************************************************
 * Name : TaskIsOpened
 * Desc : see if task is opened
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * Out  : 
 * Auth : lb, 01.08.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 ***************************************************************************/
Boolean TaskIsOpened(DmOpenRef dbP, UInt16 index)
{
	TaskAttrType attr;

	DBGMSG((DBB, "TaskIsOpened"));

	if(!TaskGetAttr(dbP, index, &attr))
		return 0;
	else
		return attr.opened;
} // Boolean TaskIsOpened(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskIsVisible
 * Desc : true if task is first level, or in an opened hierarchy
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * Out  : 
 * Auth : lb, 02.08.2000
 * Mod  : lb, 24.08.2000
 * 			don't look beyond gRefLevel
 * 		  lb, 2001-09-06
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
Boolean TaskIsVisible(DmOpenRef dbP, UInt16 index)
{
	DBGMSG((DBB, "TaskIsVisible"));

	// hide done tasks
	if (gProjectPrefs.hideDoneTasks && (TaskGetCompleted(dbP, index) == 10))
		return false;

	while (TaskGetLevel(dbP, index) > gRefLevel + 1)
	{
		index = TaskGetFatherIndex(dbP, index);
		if (!TaskIsOpened(dbP, index))
			return false;
	}

	return true;
} // Boolean TaskIsVisible(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskMakeVisible
 * Desc : open all this task's fathers
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * Out  : 
 * Auth : lb, 19.10.2000
 * Mod	: lb, 2001-09-06
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
void TaskMakeVisible(DmOpenRef dbP, UInt16 index)
{
	DBGMSG((DBB, "TaskMakeVisible"));

	// hide done tasks on => nothing to do, the task is hidden
	if (gProjectPrefs.hideDoneTasks && (TaskGetCompleted(dbP, index) == 10))
		return;

	while (TaskGetLevel(dbP, index) > gRefLevel + 1)
	{
		index = TaskGetFatherIndex(dbP, index);
		if (!TaskIsOpened(dbP, index))
			TaskSetOpened(dbP, index, true);
	}
} // void TaskMakeVisible(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskToggleIsOpened
 * Desc : toggle isOpened flag
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * Out  : 
 * Auth : lb, 01.08.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 ***************************************************************************/
pgErr TaskToggleIsOpened(DmOpenRef dbP, UInt16 index)
{
	TaskAttrType attr;

	DBGMSG((DBB, "TaskToggleIsOpened"));

	if(!TaskGetAttr(dbP, index, &attr))
		return pgError;
	attr.opened = !attr.opened;
	TaskSetAttr(dbP, index, attr);
	return pgOK;
} // pgErr TaskToggleIsOpened(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSetOpened
 * Desc : set isOpened flag
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * Out  : 
 * Auth : lb, 04.08.2000
 * Mod  : lb, 22.08.2000
 * 			new preferences scheme
 * Rem  : bad err handling
 * TODO : better error handling
 ***************************************************************************/
pgErr TaskSetOpened(DmOpenRef dbP, UInt16 index, Boolean value)
{
	TaskAttrType attr;

	DBGMSG((DBB, "TaskSetOpened"));

	if(!TaskGetAttr(dbP, index, &attr))
		return pgError;
	attr.opened = value;
	TaskSetAttr(dbP, index, attr);

	return pgOK;
} // pgErr TaskSetOpened(DmOpenRef dbP, UInt16 index, Boolean value)



/****************************************************************************
 * Name : TaskGetFatherIndex
 * Desc : get the father of a task
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * Out  : father, dmMaxRecordIndex if not found
 * Auth : lb, 28.07.2000
 * Rem  : bad err handling
 * TODO : better error handling
 ***************************************************************************/
UInt16 TaskGetFatherIndex(DmOpenRef dbP, UInt16 index)
{
	UInt16 prev = 0;
	UInt16 sindex;
	UInt8 level;

	DBGMSG((DBB, "TaskGetFatherIndex"));

	level = TaskGetLevel(dbP, index);

	// top level tasks have no father
	if (level == FIRST_LEVEL)
		return prev;

	sindex = index - 1;

	while (sindex < index)
	{
		if (TaskGetLevel(dbP, sindex) == level - 1)
		{
			prev = sindex;
			break;
		}
		sindex--;
	}
	return prev;
} // UInt16 TaskGetFatherIndex(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskGetPrevIndex
 * Desc : get the previous of a task
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * Out  : previous, dmMaxRecordIndex if not found
 * Auth : lb, 27.07.2000
 * Rem  : bad err handling
 * TODO : better error handling
 ***************************************************************************/
UInt16 TaskGetPrevIndex(DmOpenRef dbP, UInt16 index)
{
	UInt16 prev = dmMaxRecordIndex;
	UInt16 sindex;
	UInt8  level;
	UInt8  actLevel;

	DBGMSG((DBB, "TaskGetPrevIndex"));

	if (!TaskGetHasPrev(dbP, index))
		return prev;

	level = TaskGetLevel(dbP, index);
	sindex = index - 1;

	while (sindex < index)
	{
		// get level of item sindex
		actLevel = TaskGetLevel(dbP, sindex);

		// if it's the same, we found
		if (actLevel == level)
		{
			prev = sindex;
			break;
		}

		// if it's less, there's no previous
		if (actLevel < level)
		{
			break;
		}
		sindex--;
	}
	return prev;
} // UInt16 TaskGetPrevIndex(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskGetPrevIndexByLevel
 * Desc : get the previous of same level from index
 * Parm : 
 * 			-> db to search
 * 			-> index of the record
 * 			-> level to search for
 * Out  : previous, dmMaxRecordIndex if not found
 * Auth : lb, 31.07.2000
 * Rem  : bad err handling
 * TODO : better error handling
 ***************************************************************************/
UInt16 TaskGetPrevIndexByLevel(DmOpenRef dbP, UInt16 index, UInt8 level)
{
	UInt16 prev = dmMaxRecordIndex;
	UInt16 sindex = index - 1;
	UInt8  actLevel;

	DBGMSG((DBB, "TaskGetPrevIndexByLevel"));

	// cannot search before first item
	if (index == FIRST_INDEX)
		return prev;

	// search all preceding items
	while (sindex < index)
	{
		// get level of item sindex
		actLevel = TaskGetLevel(dbP, sindex);

		// if it's the same, we found
		if (actLevel == level)
		{
			prev = sindex;
			break;
		}

		// if it's less, there's no previous
		if (actLevel < level)
		{
			break;
		}
		sindex--;
	}

	return prev;
} // UInt16 TaskGetPrevIndexByLevel(DmOpenRef dbP, UInt16 index, UInt8 level)



/****************************************************************************
 * Name : TaskGetNextIndex
 * Desc : get the next of a task
 * Parm : 
 * 			-> index of the record
 * Out  : next, 0 if not found
 * Auth : lb, 27.07.2000
 * Mod  : 01.08.2000
 * 			corrected bug that found a next in other branchs
 * Rem  : bad err handling
 * TODO : better error handling
 ***************************************************************************/
UInt16 TaskGetNextIndex(DmOpenRef dbP, UInt16 index)
{
	UInt16 next = 0, lastRecord = PgNumRecords(dbP);
	UInt8 level;
	UInt8 actLevel;

	DBGMSG((DBB, "TaskGetNextIndex"));

	level = TaskGetLevel(dbP, index);
	index++;

	while (index < lastRecord)
	{
		actLevel = TaskGetLevel(dbP, index);
		if (actLevel == level)
		{
			next = index;
			break;
		}

		if (actLevel < level)
		{
			break;
		}
		index++;
	}
	return next;
} // UInt16 TaskGetNextIndex(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskGetNextRelativeIndex
 * Desc : get the next of a task, may be the next of a father (TODO be clearer...)
 * Parm : 
 * 			-> index of the record
 * Out  : next, 0 if not found
 * Auth : lb, 01.08.2000
 * Mod  : lb, 09.08.2000
 * 			- adding support for subtree viewing
 * 		  lb, 28.09.2000
 * 		  	- return index if index = dmMaxRecordIndex
 * Rem  : bad err handling
 * TODO : better error handling
 ***************************************************************************/
UInt16 TaskGetNextRelativeIndex(DmOpenRef dbP, UInt16 index)
{
	UInt16 next = 0, lastRecord = PgNumRecords(dbP);
	UInt8 level;
	UInt8 actLevel;

	DBGMSG((DBB, "TaskGetNextRelativeIndex"));

	if (lastRecord == 0)
		return next;
	
	if (index == dmMaxRecordIndex)
		return index;

	level = TaskGetLevel(dbP, index);
	index++;

	while (index < lastRecord)
	{
		actLevel = TaskGetLevel(dbP, index);
		//if (actLevel <= level && actLevel >= gRefLevel)
		if (actLevel <= level)
		{
			next = index;
			break;
		}

		index++;
	}
	return next;
} // UInt16 TaskGetNextRelativeIndex(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskGetType
 * Desc : get the type of a task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * Out  : ItemType type
 * Auth : lb, 2001-09-06
 ***************************************************************************/
ItemType TaskGetType(DmOpenRef dbP, UInt16 index)
{
	ItemType returnValue;

	DBGMSG((DBB, "TaskGetType"));


	// +++ FIX THIS +++
	// How about we figure out how this data structure is getting corrupt in the first place?
	// +++ FIX THIS +++
        // make sure a corrupt data structure does not crash program
	 returnValue=TaskGetFormat(dbP, index).itemType;

        if (returnValue >= invalidType)
	{
		DEBUG1("Something trashed format.bits.itemType, resetting to progressType");
		returnValue=progressType;
	}

	return returnValue;

} // ItemType TaskGetType(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSetType
 * Desc : set the type of a task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * 			-> type
 * Out  : 
 * Auth : lb, 2001-09-06
 * Rem  : Replaces TaskSetActionState, which was good at the time there were
 * 		  only progress and action types...
 * Mod	: lb, 2001-11-25
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
pgErr TaskSetType(DmOpenRef dbP, UInt16 index, ItemType type)
{
	UInt8 completed;
	TaskFormatType format;
	MemHandle h;
	TaskExtendedRecordType *p;

	DBGMSG((DBB, "TaskSetType"));

	h = DmGetRecord(dbP, index);
	p = MemHandleLock(h);
	completed = StdFields(p)->completed;
	format = p->format.bits;

	// convert the completed value when needed
	switch (type)
	{
		case actionType:
			if (completed != 10)
				completed = 0;
			break;
		case informativeType:
			completed = 0;
			break;
		default:
			break;
	}

	format.itemType = type;

	DmWrite(p, StdFieldsOffset(p) + OffsetOf(TaskStandardFields, completed), 
		&completed, sizeof(UInt8));
	DmWrite(p, OffsetOf(TaskRecordType, format), &format, 
		sizeof(TaskFormatType));
	MemHandleUnlock(h);
	DmReleaseRecord(dbP, index, true);
	return pgOK;
} // pgErr TaskSetType(DmOpenRef dbP, UInt16 index, ItemType type)



/****************************************************************************
 * Name : TaskGetCompleted
 * Desc : get completed value of a task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * Out  : 
 * Auth : lb, 04.08.2000
 * Mod	: lb, 2001-09-09
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
UInt8 TaskGetCompleted(DmOpenRef dbP, UInt16 index)
{
	MemHandle h;
	TaskExtendedRecordType *p;
	UInt8 value;

	DBGMSG((DBB, "TaskGetCompleted"));

	h = DmQueryRecord(dbP, index);
	p = MemHandleLock(h);
	value = StdFields(p)->completed;
	MemHandleUnlock(h);
	return value;
} // UInt8 TaskGetCompleted(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSetCompleted
 * Desc : set completed value of a task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * 			-> value
 * 			-> need to set the dirty bit ?
 * Out  : 
 * Auth : lb, 04.08.2000
 * Mod  : lb, 08.05.2001
 * 			- don't change the task's type
 * 			- adapt to ACTION type
 * 			=> same day : was a very bad idea, undo
 * Mod	: lb, 2001-09-09
 * 		  	adapt to 0.23 db format
 * 		  lb, 2002-01-21
 * 		  	added dirty param, because TaskCalcCompleted must not set dirty
 ***************************************************************************/
pgErr TaskSetCompleted(DmOpenRef dbP, UInt16 index, UInt8 value, Boolean dirty)
{
	MemHandle h;
	TaskExtendedRecordType *p;

	DBGMSG((DBB, "TaskSetCompleted"));

	h = DmGetRecord(dbP, index);
	p = MemHandleLock(h);
	DmWrite(p, StdFieldsOffset(p) + OffsetOf(TaskStandardFields, completed), 
		&value, sizeof(UInt8));
	MemHandleUnlock(h);
	DmReleaseRecord(dbP, index, dirty);
	return pgOK;
} // pgErr TaskSetCompleted(DmOpenRef dbP, UInt16 index, UInt8 value, 
  // Boolean dirty)



/****************************************************************************
 * Name : TaskGetIsDone
 * Desc : true if the task is 100% done or ACTION_OK
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * Out  : true if the task is 100% done or ACTION_OK
 * Auth : lb, 09.08.2000
 * Rem  : bad err handling
 * Mod	: lb, 2001-09-06
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
Boolean TaskGetIsDone(DmOpenRef dbP, UInt16 index)
{
	DBGMSG((DBB, "TaskGetIsDone"));

	return (TaskGetCompleted(dbP, index) == 10);
} // Boolean TaskGetIsDone(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskGetActionState
 * Desc : get action state of a task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * Out  : action status
 * Auth : lb, 03.08.2000
 * Rem  : bad err handling
 * TODO : better error handling
 * Mod  : lb, 2001-08-23
 * 			adapted for info types
 ***************************************************************************/
Boolean TaskGetActionState(DmOpenRef dbP, UInt16 index)
{
	DBGMSG((DBB, "TaskSetActionState"));

	return TaskGetFormat(dbP, index).itemType == actionType;
} // Boolean TaskGetActionState(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSetDueDate
 * Desc : set the due date of a task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * 			-> date
 * Out  : 
 * Auth : lb, 06.08.2000
 * Mod	: lb, 2001-09-09
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
pgErr TaskSetDueDate(DmOpenRef dbP, UInt16 index, DateType date)
{
	MemHandle h;
	TaskExtendedRecordType *p;

	DBGMSG((DBB, "TaskSetDueDate"));

	h = DmGetRecord(dbP, index);
	if (h)
	{
		p = MemHandleLock(h);
		if (p)
			DmWrite(p, 
				StdFieldsOffset(p) + OffsetOf(TaskStandardFields, dueDate), 
				&date, sizeof(DateType));
		MemHandleUnlock(h);
	}

	DmReleaseRecord(dbP, index, true);
	return pgOK;
} // pgErr TaskSetDueDate(DmOpenRef dbP, UInt16 index, DateType date)



/****************************************************************************
 * Name : TaskGetDueDate
 * Desc : get the due date of a task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * Out  : 
 * Auth : lb, 03.08.2000
 * Mod	: lb, 2001-09-09
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
DateType TaskGetDueDate(DmOpenRef dbP, UInt16 index)
{
	MemHandle h;
	TaskExtendedRecordType *p;
	DateType date;

	DBGMSG((DBB, "TaskGetDueDate"));

	h = DmQueryRecord(dbP, index);
	if (h)
	{
		p = MemHandleLock(h);
		if (p)
			date = StdFields(p)->dueDate;
		MemHandleUnlock(h);
	}

	return date;
} // void TaskGetDueDate(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSetNote
 * Desc : set the note of a task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * 			-> pointer to the note (will be copied), "" to remove the note
 * Out  : 
 * Auth : lb, 22.08.2000
 * Mod	: lb, 2001-09-09
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
pgErr TaskSetNote(DmOpenRef dbP, UInt16 index, Char* note)
{
	MemHandle h;
	TaskExtendedRecordType *p;
	UInt16 newSize, noteSize, descSize;
	TaskFormatType format;
	Char* desc;

	DBGMSG((DBB, "TaskSetNote"));

	noteSize = StrLen(note);

	// TODO : this will crash when trying to add a note to a direct link
	desc = TaskGetDescription(dbP, index);
	descSize = StrLen(desc);
	MemPtrFree(desc);
	desc=NULL;

	h = DmGetRecord(dbP, index);
	newSize = MemHandleSize(h) - StrLen(TaskGetNote(dbP, index)) + noteSize;

	h = DmResizeRecord(dbP, index, newSize);
	p = MemHandleLock(h);

	DmStrCopy(p, 
		StdFieldsOffset(p) + OffsetOf(TaskStandardFields, description) +
		descSize + 1, note);

	format = p->format.bits;
	format.hasNote = noteSize ? 1 : 0;

	DmWrite(p, OffsetOf(TaskExtendedRecordType, format), &format, 
		sizeof(TaskFormatType));

	MemHandleUnlock(h);

	DmReleaseRecord(dbP, index, true);
	return pgOK;
} // pgErr TaskSetNote(DmOpenRef dbP, UInt16 index, Char* note)



/****************************************************************************
 * Name : TaskSetDescription
 * Desc : set the description of a task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * 			-> pointer to the description (will be copied)
 * Out  : 
 * Auth : lb, 2001-09-09
 * Mod	: lb, 2001-11-25
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
pgErr TaskSetDescription(DmOpenRef dbP, UInt16 index, Char* desc)
{
	MemHandle h;
	TaskExtendedRecordType *p;
	UInt16 newSize, noteSize, newDescSize, oldSize, oldDescSize;
	Char *oldDesc = NULL;

	DBGMSG((DBB, "TaskSetDescription"));

	h = DmGetRecord(dbP, index);
	noteSize = StrLen(TaskGetNote(dbP, index));
	oldSize = MemHandleSize(h);
	newDescSize = StrLen(desc);
	oldDesc = TaskGetDescription(dbP, index);
	oldDescSize = StrLen(oldDesc);
	MemPtrFree(oldDesc);
	newSize = oldSize - oldDescSize + newDescSize;
	DBGMSG((DBB, "oldSize : %hu", oldSize));
	DBGMSG((DBB, "noteSize : %hu", noteSize));
	DBGMSG((DBB, "newDescSize : %hu", newDescSize));
	DBGMSG((DBB, "oldDescSize : %hu", oldDescSize));
	DBGMSG((DBB, "newSize : %hu", newSize));

	DBGMSG((DBB, "Before resizing the record"));
	dbgPrintTask(dbP, index);

	if (newSize > oldSize)
	{
		h = DmResizeRecord(dbP, index, newSize);
		if (!h)
		{
			MessageBox(StrNoMemorySpaces);
			return pgError;
		}
	}

	p = MemHandleLock(h);

	DBGMSG((DBB, "After resize, before moving the note"));
	dbgPrintTask(dbP, index);

	// move the note
	if (newSize != oldSize)
	{
		Char *oldNote = NULL;
		Char *oldNoteBak = NULL;
		oldNoteBak = MemPtrNew(noteSize + 1);
		oldNote = TaskGetNote(dbP, index); // buffer is not locked, get it again
		SafeCopy(oldNoteBak,oldNote,noteSize+1);
		DmWrite(p, 
			StdFieldsOffset(p) + OffsetOf(TaskStandardFields, description)
			+ newDescSize + 1, oldNoteBak, noteSize + 1);

//		DBGMSG((DBB,"Moved note"));
//		DBGMSG((DBB,"old note [%s]",oldNoteBak));
//		DBGMSG((DBB,"Len of moved note : %d", noteSize));
		MemPtrFree(oldNoteBak);
	}

	DBGMSG((DBB, "After moving the note"));
	dbgPrintTask(dbP, index);

	// write description
	DmStrCopy(p, 
		StdFieldsOffset(p) + OffsetOf(TaskStandardFields, description), desc);

	DBGMSG((DBB, "After writing the description"));
	dbgPrintTask(dbP, index);

	if (oldSize > newSize)
	{
		h = DmResizeRecord(dbP, index, newSize);
	}

	DBGMSG((DBB, "After final resize"));
	dbgPrintTask(dbP, index);

	MemHandleUnlock(h);
	DmReleaseRecord(dbP, index, true);
	DBGMSG((DBB,"TaskSetDesc OK"));

	return pgOK;
} // pgErr TaskSetDescription(DmOpenRef dbP, UInt16 index, Char* desc)



/****************************************************************************
 * Name : TaskGetDescriptionEmpty
 * Desc : get the description of a task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * Out  : Boolean
 * 
 * Auth : rp, 13.08.2003
 * This is a very inneficient way to find out if the description is empty,
 * however, the code to get the description is very complex, and it is not
 * a good idea to duplicate it.
 *
 ***************************************************************************/
Boolean TaskGetDescriptionEmpty(DmOpenRef dbP, UInt16 index)
{
	char* desc = NULL;
	Boolean descriptionIsEmpty;

	DBGMSG((DBB, "TaskGetDescriptionEmpty"));

	desc = TaskGetDescription(dbP,index);
	if (*desc)
		descriptionIsEmpty=false;
	else
		descriptionIsEmpty=true;

	MemPtrFree(desc);
	
	return descriptionIsEmpty;
} // Char* TaskGetDescription(DmOpenRef dbP, UInt16 index)





/****************************************************************************
 * Name : TaskGetDescription
 * Desc : get the description of a task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * Out  : pointer to the desc, use it immediately, it could becomme invalid
 * 		  reading only !!!
 * Auth : lb, 22.09.2000
 *        seagull, 9.10.2000
 *           - allocate tmp buffer (because unlocked ptr make warnning
 *             under POSE)
 *           - support direct linked task.
 ***************************************************************************/
Char* TaskGetDescription(DmOpenRef dbP, UInt16 index)
{
	char* desc = NULL;
	MemHandle h = NULL;
	TaskExtendedRecordType* pTask;

	DBGMSG((DBB, "TaskGetDescription"));

	h = DmQueryRecord(dbP, index);
	if (h)
	{
		pTask = MemHandleLock(h);
		if (pTask)
		{
			desc = TaskGetDescriptionByTaskPtr(pTask);
			MemHandleUnlock(h);
		}
	}
	return desc;
} // Char* TaskGetDescription(DmOpenRef dbP, UInt16 index)





/****************************************************************************
 * Name : TaskGetDescriptionByTaskPtr
 * Desc : get the description of a task
 * Parm : 
 * 			-> Task pointer
 * Out  : pointer to the desc. must be freed by the caller.
 * Auth : seagull, 9.10.2000
 * 		  burgbach, 2001-08-03
 * 		  	- added first name to the address desc
 * Mod	: lb, 2001-09-10
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
Char* TaskGetDescriptionByTaskPtr(TaskExtendedRecordType* pTask)
{
	DmOpenRef db = NULL;
	MemHandle hLink = NULL;
	Char* desc = NULL;
	Char* result = NULL;
	Boolean freeDesc = 0;

	DBGMSG((DBB, "TaskGetDescriptionByTaskPtr"));
// +++ FIX THIS +++
// This code is probably no longer needed since we don't support links anymore
// +++ FIX THIS +++
	if (pTask->format.bits.itemType == linkType)
	{
		UInt16 index;
		UInt16 attr;
		AddrDBRecordType item;
		
		if (pTask->fields.link.builtin)
		{
			if (pTask->fields.link.dbID == memoLink)
			{
				if (! MemoGetDatabase(&db, dmModeReadOnly) &&
					! DmFindRecordByID(db, pTask->fields.link.uniqueID, &index) &&
					! DmRecordInfo(db, index, &attr, NULL, NULL) &&
					! (attr & dmRecAttrDelete)  )
				{
					hLink = DmQueryRecord(db, index);
					desc = MemHandleLock(hLink);
				}
			}
			else if (pTask->fields.link.dbID == addressLink)
			{
				if (! AddrGetDatabase(&db, dmModeReadWrite) &&
					! DmFindRecordByID(db, pTask->fields.link.uniqueID, &index) &&
					! DmRecordInfo(db, index, &attr, NULL, NULL) &&
					! (attr & dmRecAttrDelete)  )
				{
					UInt16 lenNeeded = 0;
					AddrGetRecord(db, index, &item, &hLink);
					if (item.fields[name])
						lenNeeded = StrLen(item.fields[name]);
					if (item.fields[firstName])
						lenNeeded += StrLen(item.fields[firstName]);
					desc = MemPtrNew(lenNeeded + 2 + 1);
					desc[0]=chrNull;
					if (item.fields[name])
						StrCat(desc, item.fields[name]);
					if (item.fields[name] && item.fields[firstName])
						StrCat(desc, ", ");
					if (item.fields[firstName])
						StrCat(desc, item.fields[firstName]);
					freeDesc = 1;
				}
			}
		}
		else // not builtin
		{
			// TODO : complete for not builtin link
		}
	}
	else
	{
		desc = &(StdFields(pTask)->description);
	}
	
	if (desc)
	{
		UInt16 len;
		Char* pp = desc;
		// because the title of a memo ends with a \n
		while (*pp && *pp != '\0' && *pp != '\n') pp++;
		len = pp - desc;
		
		result = MemPtrNew(len + 1);
		MemMove(result, desc, len);
		result[len] = '\0';
	}
	
	if (db) DmCloseDatabase(db);
	if (hLink) MemHandleUnlock(hLink);
	if (freeDesc) MemPtrFree(desc);

	return result;
} // Char* TaskGetDescriptionByTaskPtr(TaskExtendedRecordType* pTask)



/****************************************************************************
 * Name : TaskGetNote
 * Desc : get the note of a task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * Out  : pointer to the note, use it immediately, it could becomme invalid
 * 		  reading only !!!
 * Auth : lb, 22.08.2000
 * Mod	: lb, 2001-09-09
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
Char* TaskGetNote(DmOpenRef dbP, UInt16 index)
{
	MemHandle h;
	TaskExtendedRecordType *p;
	Char* note = NULL;

	DBGMSG((DBB, "TaskGetNote"));

	h = DmQueryRecord(dbP, index);
	if (h)
	{
		p = MemHandleLock(h);
		if (!p)
			return NULL;
		note = &StdFields(p)->description;
		// go to the note field
		note += StrLen(note) + 1;
		if (note > (Char*)p + MemHandleSize(h))
		{
			note = NULL;
		}
		MemHandleUnlock(h);
	}

	DBGMSG((DBB, "TaskGetNote Exit"));

	return note;
} // Char* TaskGetNote(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskGetToDoUniqueID
 * Desc : get the uniqueID of the sync ToDo task
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * Out  : uniqueID
 * Auth : lb, 31.08.2000
 ***************************************************************************/
UInt32 TaskGetToDoUniqueID(DmOpenRef dbP, UInt16 index)
{
	UInt32 uniqueID = 0;
	UInt16 size = sizeof(uniqueID);

	DBGMSG((DBB, "TaskGetToDoUniqueID"));

	if (! TaskGetFormat(dbP, index).hasToDo)
		return 0;

	TaskGetExtraChunk(dbP, index, Extra_Link_ToDo, 0xff, 
		(MemPtr)(&uniqueID), &size);

	return uniqueID;
} // UInt32 TaskGetToDoUniqueID(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskInputDueDate
 * Desc : ask user for a due date
 * Parm : 
 * 			-> database pointer
 * 			-> index of the record
 * Out  : 
 * Auth : lb, 03.08.2000
 * Rem  : bad err handling
 * TODO : better error handling
 ***************************************************************************/
pgErr TaskInputDueDate(DmOpenRef dbP, UInt16 index)
{
	//TaskRecordType *p;
	//MemHandle h;
	DateType date;
	Boolean selected;
	UInt16 month, day, year;

	DBGMSG((DBB, "TaskInputDueDate"));

	date = TaskGetDueDate(dbP, index);
	if (date.month == 0)
	{
		date = gToday;
	}

	month = date.month;
	day = date.day;
	year = date.year + YEAR_OFFSET; // year offset
	selected = SelectDay(selectDayByDay, &month, &day, &year, "Choose a due date");
	if (selected)
	{
		date.month = month;
		date.day = day;
		date.year = year - YEAR_OFFSET;
		TaskSetDueDate(dbP, index, date);
		// This caused a bug with the new db format of 0.23
		//h = DmGetRecord(dbP, index);
		//p = MemHandleLock(h);
		//DmWrite(p, OffsetOf(TaskRecordType, dueDate), &date, sizeof(DateType));
		//MemHandleUnlock(h);
		//DmReleaseRecord(dbP, index, true);
	}
	return pgOK;
} // pgErr TaskInputDueDate(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSave
 * Desc : (temporarily) save a record (copy it and create as last index)
 * In   : 
 * 			-> record to save
 * Out  : 
 * Auth : lb, 03.08.2000
 * Rem  : doesn't handle err codes
 * Mod  : lb, 07.09.2000
 * 			- keep the same uniqueID and attr
 ***************************************************************************/
pgErr TaskSave(UInt16 index)
{
	MemHandle hNew, hOld;
	TaskRecordType *pNew, *pOld;
	UInt32 size = TaskRecordTypeSize, uniqueID;
	UInt16 realIndex = dmMaxRecordIndex, attr;

	DBGMSG((DBB, "TaskSave"));

	// get a handle on the record to save
	hOld = DmQueryRecord(gdbP, index);

	// calc the needed size
	size = MemHandleSize(hOld);

	// get a handle on a newly created record
	hNew = DmNewRecord(gdbP, &realIndex, size);

	// get a pointer on the record
	pNew = MemHandleLock(hNew);

	// get a pointer on it
	pOld = MemHandleLock(hOld);

	// write the data
	DmWrite(pNew, 0, pOld, size);

	// unlock the pointers
	MemHandleUnlock(hNew);
	MemHandleUnlock(hOld);

	// unlock the record
	DmReleaseRecord(gdbP, realIndex, true);

	// to allow sync
	// get the unique ID of original task
	DmRecordInfo(gdbP, index, &attr, &uniqueID, NULL);
	// set the unique ID of the new task
	DmSetRecordInfo(gdbP, realIndex, &attr, &uniqueID);
	return pgOK;
} // pgErr TaskSave(UInt16 index)



/****************************************************************************
 * Name : TaskGetSaved
 * Desc : get the (temporarily) saved record (copy from last index)
 * In   : 
 * 			-> index to put saved record
 * Out  : 
 * Auth : lb, 03.08.2000
 * Rem  : doesn't handle err codes
 * Mod  : lb, 07.09.2000
 * 			- keep the same uniqueID and attr
 ***************************************************************************/
pgErr TaskGetSaved(UInt16 index)
{
	MemHandle hNew, hOld;
	TaskRecordType *pNew, *pOld;
	UInt32 size = TaskRecordTypeSize, uniqueID;
	UInt16 attr;

	DBGMSG((DBB, "TaskGetSaved"));

	// get a handle on the saved record
	hOld = DmQueryRecord(gdbP, DmNumRecords(gdbP) - 1);

	// calc the needed size
	size = MemHandleSize(hOld);

	// get a handle on a newly created record
	hNew = DmNewRecord(gdbP, &index, size);

	// get a pointer on the record
	pNew = MemHandleLock(hNew);

	// get a pointer on it
	pOld = MemHandleLock(hOld);

	// write the data
	DmWrite(pNew, 0, pOld, size);

	// unlock the pointers
	MemHandleUnlock(hNew);
	MemHandleUnlock(hOld);

	// unlock the record
	DmReleaseRecord(gdbP, index, true);

	// to allow sync
	// get the unique ID of original task
	DmRecordInfo(gdbP, DmNumRecords(gdbP) - 1, &attr, &uniqueID, NULL);

	// set the unique ID of the new task
	DmSetRecordInfo(gdbP, index, &attr, &uniqueID);

	// remove saved record
	DmRemoveRecord(gdbP, DmNumRecords(gdbP) - 1);

	// remove replaced record
	DmRemoveRecord(gdbP, index + 1);

	return pgOK;
} // pgErr TaskGetSaved(UInt16 index)



/****************************************************************************
 * Name : TaskRemoveSaved
 * Desc : remove the (temporarily) saved record
 * In   : 
 * Out  : 
 * Auth : lb, 03.08.2000
 * Rem  : doesn't handle err codes
 ***************************************************************************/
pgErr TaskRemoveSaved(void)
{
	DBGMSG((DBB, "TaskRemoveSaved"));

	// remove replaced record
	DmRemoveRecord(gdbP, DmNumRecords(gdbP) - 1);
	return pgOK;
} // pgErr TaskRemoveSaved(void)



/****************************************************************************
 * Name : TaskRemove
 * Desc : remove a task and all its children
 * In   : 
 * 			-> index of the task to remove
 * Out  : 
 * Auth : lb, 03.08.2000
 * Rem  : doesn't handle err codes
 * Mod  : lb, 2001-01-21
 * 			call DmRemoveRecord instead of DmDeleteRecord
 ***************************************************************************/
pgErr TaskRemove(DmOpenRef dbP, UInt16 index)
{
	Boolean stillChild = false;

	DBGMSG((DBB, "TaskRemove"));

	// as long as I have a child, remove it
	while (TaskGetHasChild(dbP, index))
	{
		TaskRemove(dbP, index + 1);
	}

	if (TaskGetHasNext(dbP, index))
	{
		if (!TaskGetHasPrev(dbP, index))
		{
			TaskSetHasPrev(dbP, index + 1, false);
		}
		stillChild = true;
	}

	if (TaskGetHasPrev(dbP, index))
	{
		if (!TaskGetHasNext(dbP, index))
		{
			TaskSetHasNext(dbP, TaskGetPrevIndex(dbP, index), false);
		}
		stillChild = true;
	}

	if (!stillChild)
	{
		if (TaskGetLevel(dbP, index) != 1)
		{
			TaskSetHasChild(dbP, TaskGetFatherIndex(dbP, index), false);
		}
	}

	// then, remove me
	DmRemoveRecord(dbP, index);

	// we used this code in 0.23, but it's not needed to keep the record
	// for the hotsync, so I prefer to remove the record, not to clutter the
	// database with 0-length records
	//DmDeleteRecord(dbP, index);
	//DmMoveRecord(dbP, index, DmNumRecords(dbP));
	//gNumDeleted++;
	return pgOK;
} // pgErr TaskRemove(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskRemoveDone
 * Desc : remove all done tasks
 * In   : 
 * Out  : 
 * Auth : lb, 29.08.2000
 * Rem  : doesn't handle err codes
 ***************************************************************************/
pgErr TaskRemoveChildren(DmOpenRef dbP, UInt16 index)
{
	DBGMSG((DBB, "TaskRemoveChildren"));

	// as long as I have a child, remove it
	while (TaskGetHasChild(dbP, index))
	{
		TaskRemove(dbP, index + 1);
	}

	TaskSetHasChild(gdbP, index, false);

	return pgOK;
} // pgErr TaskRemoveChildren(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskRemoveDone
 * Desc : remove all done tasks
 * In   : 
 * Out  : 
 * Auth : lb, 29.08.2000
 * Rem  : doesn't handle err codes
 * Mod	: lb, 2001-09-06
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
pgErr TaskRemoveDone(DmOpenRef dbP)
{
	UInt16 i = 1;

	DBGMSG((DBB, "TaskRemoveDone"));

	while (i < PgNumRecords(dbP))
	{
		if (TaskGetCompleted(dbP, i) == 10)
		{
			TaskRemove(dbP, i);
		}
		else
		{
			i++;
		}
	}
	return pgOK;
} // pgErr TaskRemoveDone(DmOpenRef dbP)



/****************************************************************************
 * Name : TaskRemoveDoneChildren
 * Desc : remove all done tasks in the children of given task
 * In   : 
 * Out  : 
 * Auth : lb, 29.08.2000
 * Rem  : doesn't handle err codes
 * Mod	: lb, 2001-09-06
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
pgErr TaskRemoveDoneChildren(DmOpenRef dbP, UInt16 index)
{
	UInt16 i = index + 1, refLevel = TaskGetLevel(dbP, index);

	DBGMSG((DBB, "TaskRemoveDoneChildren"));

	while (i < (PgNumRecords(dbP)) && TaskGetLevel(dbP, i) > refLevel)
	{
		if (TaskGetCompleted(dbP, i) == 10)
		{
			TaskRemove(dbP, i);
		}
		else
		{
			i++;
		}
	}
	return pgOK;
} // pgErr TaskRemoveDoneChildren(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskRemoveDeletedRecords
 * Desc : Remove all records with the deleted flag
 * In   : -> dbP
 * Out  : 
 * Auth : lb, 2002-01-21
 ***************************************************************************/
void TaskRemoveDeletedRecords(DmOpenRef dbP)
{
	Int32 i;
	UInt16 attr;

	DBGMSG((DBB, "Starting TaskRemoveDeletedRecords"));

	i = PgNumRecords(dbP) - 1;
	while (i > 0)
	{
		DmRecordInfo(dbP, i, &attr, NULL, NULL);
		if (attr & dmRecAttrDelete)
		{
			DmRemoveRecord(dbP, i);
		}
		i--;
	}
	DBGMSG((DBB, "End of TaskRemoveDeletedRecords"));
} // void TaskRemoveDeletedRecords(DmOpenRef dbP)



/****************************************************************************
 * Name : TaskExpandAll
 * Desc : set opened flag for all tasks
 * In   : 
 * Out  : 
 * Auth : lb, 04.08.2000
 * Rem  : doesn't handle err codes
 ***************************************************************************/
pgErr TaskExpand(DmOpenRef dbP, UInt16 index)
{
	UInt16 i, num;
	UInt8 level;

	DBGMSG((DBB, "TaskExpand"));

	level = TaskGetLevel(dbP, index);
	num = PgNumRecords(dbP);
	TaskSetOpened(dbP, index, true);
	for (i = index + 1; i < num && TaskGetLevel(dbP, i) > level; i++)
	{
		TaskSetOpened(dbP, i, true);
	}
	return pgOK;
} // pgErr TaskExpand(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskCollapseAll
 * Desc : unset opened flag for all tasks
 * In   : 
 * Out  : 
 * Auth : lb, 04.08.2000
 * Rem  : doesn't handle err codes
 ***************************************************************************/
pgErr TaskCollapse(DmOpenRef dbP, UInt16 index)
{
	UInt16 i, num;
	UInt8 level;

	DBGMSG((DBB, "TaskCollapse"));

	level = TaskGetLevel(dbP, index);
	num = PgNumRecords(dbP);

	// don't collapse task 0 !!!
	if (index != 0)
	{
		TaskSetOpened(dbP, index, false);
	}

	for (i = index + 1; i < num && TaskGetLevel(dbP, i) > level; i++)
	{
		// don't collapse a task without child, or it'll be closed
		// when it becomes a child
		if (TaskGetHasChild(dbP, i))
		{
			TaskSetOpened(dbP, i, false);
		}
	}
	return pgOK;
} // pgErr TaskCollapse(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskCalcCompleted
 * Desc : calc and set completed by mean of the children
 * In   : 
 * 			-> database pointer
 * 			-> index of father to calc
 * Out  : 
 * Auth : lb, 04.08.2000
 * Mod  : hcc, 04.28.2001
 *          Preserve item's type. Ignore Informative children in calculation.
 * Mod  : lb, 2001-08-23
 * 			Fix to be able to calc an item that has no children
 * 			useful to calc after a move
 * Mod	: lb, 2001-09-06
 * 		  	adapt to 0.23 db format
 * Rem  : doesn't handle err codes
 * 		  warning : this one doesn't recurse down but UP ! You must call it
 * 		  			on a father whose children are up to date !
 ***************************************************************************/
pgErr TaskCalcCompleted(DmOpenRef dbP, UInt16 index)
{
	UInt16 mean = 0, num = 0, tmp;
	UInt16 subtask;
	UInt8  level;
	ItemType type = TaskGetType(dbP, index);
	UInt8 myCompleted;

	DBGMSG((DBB, "TaskCalcCompleted"));

	/* If I am an informative task, don't change my type. */
	if (type == informativeType)
		return pgOK;

	level = TaskGetLevel(dbP, index);

	// nothing to calc for a task without children
	// nothing to calc if index isn't a progress type
	if (type == progressType && TaskGetHasChild(dbP, index))
	{
		subtask = index + 1;
		while (subtask)
		{
			tmp = TaskGetCompleted(dbP, subtask);
			
			// don't count informative type
			if (TaskGetType(dbP, subtask) != informativeType)
			{
				mean += tmp;
				num++;
			}

			// this one will return 0 if no brother is found
			subtask = TaskGetNextIndex(dbP, subtask);
		}

		if (num != 0) // NOTE: Do nothing if all children are informative
		{
			myCompleted = mean / num;
			TaskSetCompleted(dbP, index, myCompleted, false);
		}			
	}

	// recurse up, calc for the elders
	if (level > 1)
	{
		TaskCalcCompleted(dbP, TaskGetFatherIndex(dbP, index));
	}
	return pgOK;

} // pgErr TaskCalcCompleted(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskCalcCompletedAll
 * Desc : calc and set completed by mean of the children for all fathers
 * In   : 
 * 			-> database pointer
 * Out  : 
 * Auth : lb, 04.08.2000
 * Rem  : doesn't handle err codes
 * 		  warning : this one can be lengthy !!!!
 * TODO : optimize it !!!
 ***************************************************************************/
pgErr TaskCalcCompletedAll(DmOpenRef dbP)
{
	UInt16 i;

	DBGMSG((DBB, "TaskCalcCompletedAll"));

	// children always comes after parents !
	for (i = PgNumRecords(dbP) - 1; i > 0; i--)
	{
		TaskCalcCompleted(dbP, i); // this one controls if i has a child
	}
	return pgOK;

} // pgErr TaskCalcCompletedAll(DmOpenRef dbP)



/****************************************************************************
 * Name : TaskToRight
 * Desc : task becomes a child of previous one
 * Prm  : index of future child
 * Out  : 
 * Auth : lb, 27.07.2000
 * Mod  : lb, 28.07.2000
 * 			new method, work by index
 * 		  lb, 24.08.2000
 * 		  	fixed a bug when moving right the last task
 * 		  lb, 2001-08-23
 * 		  	calc completed after move
 ***************************************************************************/
pgErr TaskToRight(DmOpenRef dbP, UInt16 index)
{
	UInt16 i;
	UInt8 actualLevel;
	UInt16 actual;
	UInt16 numRec = PgNumRecords(dbP);

	DBGMSG((DBB, "TaskToRight"));

	// can't push the first one, he must be a first level one
	if (index == FIRST_INDEX)
		return pgError;

	// can't push more than previous level + 1
	if (TaskGetLevel(dbP, index - 1) < TaskGetLevel(dbP, index))
		return pgError;

	// if actual has no next, modify previous
	if (!TaskGetHasNext(dbP, index))
	{
		TaskSetHasNext(dbP, actual = TaskGetPrevIndex(dbP, index), false);
	}

	// actual loose his next
	TaskSetHasNext(dbP, index, false);

	// inc level
	TaskSetLevel(dbP, index, actualLevel = TaskGetLevel(dbP, index) + 1);

	// if a previous exist
	if (TaskGetPrevIndex(dbP, index) != dmMaxRecordIndex)
	{
		// tell him he has a next
		TaskSetHasNext(dbP, TaskGetPrevIndex(dbP, index), true);
		// actual has a prev
		TaskSetHasPrev(dbP, index, true);
	}
	else
	{
		// no previous, actual is a first child
		// update father
		TaskSetHasChild(dbP, TaskGetFatherIndex(dbP, index), true);
		// actual has no previous
		TaskSetHasPrev(dbP, index, false);
	}

	// need to update the children
	if (TaskGetHasChild(dbP, index))
	{
		i = index + 1;
		// all children go to right without other modification
		while (i < numRec && TaskGetLevel(dbP, i) >= actualLevel)
		{
			TaskSetLevel(dbP, i, TaskGetLevel(dbP, i) + 1);
			i++;
		}
	}
	TaskCalcCompleted(dbP, index);
	return pgOK;
} // pgErr TaskToRight(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskToLeft
 * Desc : move task index to the left
 * Prm  : index of item to be pushed to left
 * Out  : 
 * Auth : lb, 31.07.2000
 * Rem  : Border effect : sets gActualTask to the new index !!!
 * Mod  : lb, 09.08.2000
 * 			- update gActualTask for continuous moving
 * 		  lb, 24.08.2000
 * 		  	fixed a bug when moving right the last task
 * 		  lb, 2001-08-23
 * 		  	calc completed after move
 ***************************************************************************/
pgErr TaskToLeft(DmOpenRef dbP, UInt16 index)
{
	UInt16 i;
	UInt8 actualLevel;
	UInt16 prev;
	UInt16 father;
	UInt16 newIndex = index;
	Boolean needMove = false;
	UInt16 numRec = PgNumRecords(dbP);

	DBGMSG((DBB, "TaskToLeft"));

	// can't push the first one, he must be a first level one
	if (index == FIRST_INDEX)
		return pgError;

	// can't push more than level 1
	if (TaskGetLevel(dbP, index) == FIRST_LEVEL)
		return pgError;

	// to gain some time
	father = TaskGetFatherIndex(dbP, index);

	// return if I can't find the father or if it's the root
	if (!father)
		return pgError;

	// if previous exists, update its next
	if ((prev = TaskGetPrevIndex(dbP, index)) != dmMaxRecordIndex)
	{
		TaskSetHasNext(dbP, prev, TaskGetHasNext(dbP, index));
	}
	else
	{
		// no previous, actual is a first child
		// update father
		// father still has child if actual has a next
		TaskSetHasChild(dbP, father, TaskGetHasNext(dbP, index));
	}

	// if actual has a next, next takes actual's prev status
	if (TaskGetHasNext(dbP, index))
	{
		TaskSetHasPrev(dbP, TaskGetNextIndex(dbP, index), TaskGetHasPrev(dbP, index));
		// actual must be moved outside its parent's scope
		// that's before father's next
		needMove = true;
	}

	// if father has no next, modify actual
	if (!TaskGetHasNext(dbP, father))
	{
		TaskSetHasNext(dbP, index, false);
	}
	else
	{
		TaskSetHasNext(dbP, index, true);
	}

	TaskSetHasPrev(dbP, index, true);

	if (needMove)
	{
		newIndex = TaskGetNextRelativeIndex(dbP, father);
		if (newIndex == 0)
		{
			newIndex = numRec;
		}
		DmMoveRecord(dbP, index, newIndex);
		// we insert after index, so newIndex becomes newIndex-- after insertion
		newIndex--;
		gActualTask = newIndex;
	}

	// dec level
	TaskSetLevel(dbP, newIndex, actualLevel = TaskGetLevel(dbP, newIndex) - 1);

	// father gain a next
	TaskSetHasNext(dbP, father, true);

	// need to update the children
	if (TaskGetHasChild(dbP, newIndex))
	{
		newIndex++; // to insert children after actual
		i = index + 1;
		// if actual has moved, first child is now index
		if (needMove)
		{
			i--;
		}

		// all children go to left, without other modification
		while (i < numRec && TaskGetLevel(dbP, i) >= actualLevel + 2)
		{
			TaskSetLevel(dbP, i, TaskGetLevel(dbP, i) - 1);
			if (needMove)
			{
				DmMoveRecord(dbP, i, newIndex);
				gActualTask--;
			}
			else
			{
				i++;
			}
		}
	}
	TaskCalcCompleted(dbP, father);
	return pgOK;
} // pgErr TaskToLeft(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskUp
 * Desc : push a task up
 * Prm  : 
 * 			-> index of task to push
 * Out  : Err code
 * Auth : lb, 01.08.2000
 * Rem  : 
 * Mod  : lb, 09.08.2000
 * 			- update gActualTask for continuous moving
 ***************************************************************************/
pgErr TaskUp(DmOpenRef dbP, UInt16 index)
{
	UInt16 prev;
	UInt8  level;
	UInt16 numRec = PgNumRecords(dbP);

	DBGMSG((DBB, "TaskUp"));

	if (!TaskGetHasPrev(dbP, index))
		return pgError;

	// update flags
	prev = TaskGetPrevIndex(dbP, index);
	TaskSetHasNext(dbP, prev, TaskGetHasNext(dbP, index));
	TaskSetHasPrev(dbP, index, TaskGetHasPrev(dbP, prev));
	TaskSetHasPrev(dbP, prev, true);
	TaskSetHasNext(dbP, index, true);

	// move index
	DmMoveRecord(dbP, index, prev);

	// update actual task
	gActualTask = prev;

	// move children
	if (TaskGetHasChild(dbP, prev))
	{
		level = TaskGetLevel(dbP, prev);

		// move children behind actual
		prev++;
		index++;

		// while it's a child
		while (index < numRec && TaskGetLevel(dbP, index) > level)
		{
			DmMoveRecord(dbP, index, prev);
			prev++;
			index++;
		}
	}
	return pgOK;
} // pgErr TaskUp(UInt16 index)



/****************************************************************************
 * Name : TaskDown
 * Desc : push a task down
 * Prm  : 
 * 			-> index of task to push
 * Out  : Err code
 * Auth : lb, 02.08.2000
 * Mod  : lb, 09.08.2000
 * 			- update gActualTask for continuous moving
 * 		  lb, 17.08.2000
 * 		    - fixed crash when moving a task that has children
 ***************************************************************************/
pgErr TaskDown(DmOpenRef dbP, UInt16 index)
{
	UInt16 next;
	UInt32 uniqueID;

	DBGMSG((DBB, "TaskDown"));

	if (!TaskGetHasNext(dbP, index))
		return pgError;

	// to locate the task after it has moved
	DmRecordInfo(dbP, gActualTask, NULL, &uniqueID, NULL);

	next = TaskGetNextIndex(dbP, index);
	TaskUp(dbP, next);

	// locate the task
	DmFindRecordByID(dbP, uniqueID, &gActualTask);

	// update the table, because if the task moved down has children, its
	// new index is not in the table, and SelectTask will crash with
	// move = true
	ProjectTableUpdate();
	return pgOK;
} // pgErr TaskDown(UInt16 index)



/****************************************************************************
 * Name : TaskPublishToDo
 * Desc : publish the actual task in the ToDo DB
 * Prm  : 
 * Out  : 
 * Auth : lb, 21.08.2000
 * Mod  : rw, 30.12.2000
 * 		  	support for syncing categories with ToDo
 * Mod	: lb, 2001-09-10
 * 		  	adapt to 0.23 db format
 * 		  	also sync the note
 ***************************************************************************/
pgErr TaskPublishToDo(DmOpenRef dbP, UInt16 index)
{
	DmOpenRef ToDoDB;
	ToDoItemType task;
	MemHandle h;
	TaskExtendedRecordType *p, *parent;
	UInt16 todoIndex = 0;
	Err err = errNone;
	UInt32 uniqueID;
	TaskFormatType format;
	Char name[dmCategoryLength];
	UInt16 category;
	#define maxDescLen 256
	Char descWithParents[maxDescLen]; // 256 from ToDoDB.c, maxDescLen

	DBGMSG((DBB, "TaskPublishToDo"));

	if (TaskGetFormat(dbP, index).hasToDo)
		return pgOK;

	// try to open the ToDo database
	if (ToDoGetDatabase(&ToDoDB, dmModeReadWrite))
	{
		FrmCustomAlert(AltEmpty, "Can't open the todo database", " ", " ");
		return pgError;
	}

	h = DmQueryRecord(dbP, index);
	if (!h)
		return pgError;

	p = MemHandleLock(h);

	// convert due date format if there's no due date
	task.dueDate = StdFields(p)->dueDate;
	if (task.dueDate.month == 0)
	{
		*((UInt16 *) &task.dueDate) = toDoNoDueDate;
	}

	// convert priority : no => 5
	task.priority = StdFields(p)->priority;
	if (!(task.priority > 0 && task.priority < NO_PRIORITY))
	{
		task.priority = 5;
	}

	if (1) // user wants complete desc ?
	{
		Char *pdesc;
		UInt16 pindex = index;
		UInt8 level = 2;
		parent = p;
		SafeCopy(descWithParents, &StdFields(p)->description,maxDescLen);
		while (level > 1)
		{
			pindex = TaskGetFatherIndex(dbP, pindex); // get father
			if (pindex == dmMaxRecordIndex) break; // check validity
			level = TaskGetLevel(dbP, pindex); // get level for stop condition
			pdesc = TaskGetDescription(dbP, pindex); // get description
			if (StrLen(pdesc) + StrLen(descWithParents) < maxDescLen - 4) // 0:
			{
				StrCat(descWithParents, " | "); // separator
				StrCat(descWithParents, pdesc);
				MemPtrFree(pdesc);
			}
			else
			{
				MemPtrFree(pdesc);
				break;
			}
		}
	}

	if (1) // complete desc
	{
		task.description = descWithParents;
	}
	else
	{
		task.description = &StdFields(p)->description;
	}
	if (TaskGetFormat(dbP, index).hasNote)
	{
		task.note = &StdFields(p)->description + 
			
				StrLen(&StdFields(p)->description) + 1;
	}
	else
	{
		task.note = "";
	}

	// match the category name from the ToDoDB
	category = TaskGetCategory(dbP, index);
	CategoryGetName(dbP, category, name);
	category = CategoryFind(ToDoDB, name);
	// Task category not in ToDo - set ToDo category to Unfiled
	if (category == dmAllCategories)
	{
		category = 0;
	}

	// create the ToDo entry
	err = ToDoNewRecord(ToDoDB, &task, category, &todoIndex);

	// Unlock the task, now that the ToDo is created
	MemHandleUnlock(h);

	if (err == errNone)
	{
		DmRecordInfo(ToDoDB, todoIndex, NULL, &uniqueID, NULL);
		DBGMSG((DBB, "ToDo UID : %ld (hex next line)", uniqueID));
		DBGMSGBIN((DBB, &uniqueID, sizeof(uniqueID)));
		TaskSetExtraChunk(dbP, index, Extra_Link_ToDo, 0x00,
						  &uniqueID, sizeof(uniqueID));

		format = TaskGetFormat(dbP, index);
		format.hasToDo = 1;
		TaskSetFormat(dbP, index, format);
	}
	else
	{
		DEBUG1("Export failed !");
	}

	DmCloseDatabase(ToDoDB);

	return pgOK;

} // pgErr TaskPublishToDo(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSetToDoStatus
 * Desc : update the sync todo
 * Prm  : 
 * 			-> database
 * 			-> index of task
 * 			-> field to update in : (one at a time)
 *				toDoPriority,
 *				toDoComplete,
 *				toDoDueDate,
 *				toDoDescription, not yet
 *				toDoNote, not yet
 *				toDoCategory,
 * 			-> value    (pass NULL if you don't want it)
 * 			-> priority (")
 * 			-> duedate  (")
 * 			-> category (")
 * Out  : 
 * Auth : lb, 30.08.2000
 * Mod  : rw, 30.12.2000
 * 		  	support for syncing categories with ToDo
 * Mod	: lb, 2001-09-06
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
pgErr TaskSetToDoStatus(DmOpenRef dbP, UInt16 index, UInt16 field, 
	UInt8 value, UInt16 priority, DateType dueDate, UInt16 category)
{
	DmOpenRef ToDoDB;
	UInt32 uniqueID;
	UInt16 todoIndex;
	UInt16 zero = 0, one = 1;
	Err err;
	Char name[dmCategoryLength];

	DBGMSG((DBB, "TaskSetToDoStatus"));

	uniqueID = TaskGetToDoUniqueID(dbP, index);

	if (!uniqueID)
		return pgError;

	// try to open the ToDo database
	if (ToDoGetDatabase(&ToDoDB, dmModeReadWrite))
	{
		FrmCustomAlert(AltEmpty, "Can't open the todo database", " ", " ");
		return pgError;
	}

	if (DmFindRecordByID(ToDoDB, uniqueID, &todoIndex))
	{
		DmCloseDatabase(ToDoDB);
		DEBUG1("The corresponding ToDo was not found, removing link.");
		TaskRemoveHasToDo(dbP, index);
		return pgError;
	}

	if (field == toDoPriority)
	{
		if (priority > 5 || priority < 1)
			priority = 5;
		err = ToDoChangeRecord(ToDoDB, &todoIndex, toDoPriority, 
			&priority);
	}

	if (field == toDoComplete)
	{
		err = ToDoChangeRecord(ToDoDB, &todoIndex, toDoComplete, 
			value == ACTION_DONE ? &one : &zero );
	}

	if (field == toDoDueDate)
	{
		if (dueDate.month == 0)
			*((UInt16 *) &dueDate) = toDoNoDueDate;

		err = ToDoChangeRecord(ToDoDB, &todoIndex, toDoDueDate, 
			(UInt16*)&dueDate);
	}

	if (field == toDoCategory)
	{
		CategoryGetName(dbP, category, name);
		category = CategoryFind(ToDoDB, name);
		// Task category not in ToDo - set ToDo category to Unfiled
		if (category == dmAllCategories)
		{
			category = 0;
		}
		err = ToDoChangeRecord(ToDoDB, &todoIndex, toDoCategory, 
							   &category);
	}

//ToDoRecordFieldType 
//typedef enum  { 
	//toDoPriority,
	//toDoComplete,
	//toDoDueDate,
	//toDoDescription,
	//toDoNote,
	//toDoCategory } ToDoRecordFieldType;
//
//#define completeFlag 	0x80
//#define priorityOnly 	~completeFlag

	DmCloseDatabase(ToDoDB);

	return pgOK;

} // pgErr TaskSetToDoStatus(DmOpenRef dbP, UInt16 index, UInt16 fields, 
// UInt8 value, UInt8 priority, DateType dueDate, UInt16 category)



/****************************************************************************
 * Name : TaskUpdateToDo
 * Desc : update the values of the ToDo
 * Prm  : 
 * 			-> database
 * 			-> index of task
 * Out  : 
 * Auth : lb, 30.08.2000
 * Mod  : rw, 30.12.2000
 * 		  	support for syncing categories with ToDo
 ***************************************************************************/
Err TaskUpdateToDo(DmOpenRef dbP, UInt16 index)
{
	Err err = errNone;

	DBGMSG((DBB, "TaskUpdateToDo"));

	// see if we're synchronized with a ToDo
	if (TaskGetFormat(dbP, index).hasToDo)
	{
		Err err;
		// check the status in ToDo and update this task
		err = TaskSetToDoStatus(dbP, index, toDoComplete,
			TaskGetCompleted(dbP, index),
			0, gNoDate, 0);
		err = TaskSetToDoStatus(dbP, index, toDoDueDate,
			0, 0,
			TaskGetDueDate(dbP, index),
			0);
		err = TaskSetToDoStatus(dbP, index, toDoPriority,
			0,
			TaskGetPriority(dbP, index),
			gNoDate, 0);
		err = TaskSetToDoStatus(dbP, index, toDoCategory,
			0, 0, gNoDate,
			TaskGetCategory(dbP, index));
	}
	return err;
} // Err TaskUpdateToDo(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskSyncAll
 * Desc : Update all the todo linked tasks
 * Prm  : 
 * Out  : 
 * Auth : lb, 30.08.2000
 * Mod  : rw, 30.12.2000
 * 		  	support for syncing categories with ToDo
 * 		  lb, 2001-08-23
 * 		  	calc father progress
 ***************************************************************************/
void TaskSyncAll(void)
{
	UInt16 numRec = PgNumRecords(gdbP);
	UInt16 i;
	DateType dueDate;
	UInt8 value, priority;
	UInt16 category;

	DBGMSG((DBB, "TaskSyncAll"));

	DBGMSG((DBB, "In TaskSyncAll (gdbp = %lu)", gdbP));
	DBGMSG((DBB, "Iterating from 1 to %hu", numRec - 1));

	for (i = 1; i < numRec; i++)
	{
		// see if we're synchronized with a ToDo
		if (TaskGetFormat(gdbP, i).hasToDo)
		{
			pgErr err;
			// check the status in ToDo and update this task
			err = TaskGetToDoStatus(gdbP, i, 
				&value, &priority, &dueDate, &category);
			if (err == pgOK)
			{
				TaskSetCompleted(gdbP, i, value, true);
				TaskSetDueDate(gdbP, i, dueDate);
				TaskSetPriority(gdbP, i, priority);
				TaskSetCategory(gdbP, i, category);
				TaskCalcCompleted(gdbP, i);
			}
		}
		
	}
} // void TaskSyncAll(void)



/****************************************************************************
 * Name : TaskRemoveHasToDo
 * Desc : remove the todo link
 * Prm  : 
 * 			-> database
 * 			-> index of task
 * Out  : 
 * Auth : lb, 30.08.2000
 * Mod  : lb, 2001-11-24
 * 		fix bugs when deleting links
 ***************************************************************************/
pgErr TaskRemoveHasToDo(DmOpenRef dbP, UInt16 index)
{
	TaskFormatType format;
	UInt32 uniqueID;
	UInt16 attr;

	DBGMSG((DBB, "TaskRemoveHasToDo"));

	// update the format field
	format = TaskGetFormat(dbP, index);
	if (format.hasToDo == 0)
		return pgOK;

	uniqueID = TaskGetToDoUniqueID(dbP, index);
	DBGMSG((DBB, "In RemoveHasToDo UID : %ld (hex next line)", uniqueID));
	DBGMSGBIN((DBB, &uniqueID, sizeof(uniqueID)));

	format.hasToDo = 0;
	TaskSetFormat(dbP, index, format);

	if (uniqueID)
	{
		DmOpenRef ToDoDB;
		UInt16 todoIndex;

		// try to open the ToDo database
		if (ToDoGetDatabase(&ToDoDB, dmModeReadWrite))
		{
			FrmCustomAlert(AltEmpty, "Can't open the todo database", " ", " ");
			return pgError;
		}

		if (DmFindRecordByID(ToDoDB, uniqueID, &todoIndex))
		{
			DmCloseDatabase(ToDoDB);
			return pgError;
		}

		DmRecordInfo(ToDoDB, todoIndex, &attr, NULL, NULL);
		if (!(attr & dmRecAttrDelete))
		{
			DmDeleteRecord(ToDoDB, todoIndex);
			DmMoveRecord(ToDoDB, todoIndex, DmNumRecords (ToDoDB));
		}

		DmCloseDatabase(ToDoDB);
	}

	TaskRemoveExtraChunk(dbP, index, Extra_Link_ToDo, 0xff);

	return pgOK;

} // pgErr TaskRemoveHasToDo(DmOpenRef dbP, UInt16 index)



/****************************************************************************
 * Name : TaskGetToDoStatus
 * Desc : check the value of this task in the ToDo database
 * Prm  : 
 * 			-> database
 * 			-> index of task
 * 			<- value    (pass NULL if you don't want it)
 * 			<- priority (")
 * 			<- duedate  (")
 * 			<- category (")
 * Out  : 
 * Auth : lb, 30.08.2000
 * Mod  : lb, 10.09.2000
 * 			- forgot to unlock the handle, fixed
 * Mod  : rw, 30.12.2000
 * 		  	support for syncing categories with ToDo
 * Mod	: lb, 2001-09-06
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
pgErr TaskGetToDoStatus(DmOpenRef dbP, UInt16 index, UInt8 *value, 
	UInt8 *priority, DateType *dueDate, UInt16 *category)
{
	DmOpenRef ToDoDB;
	MemHandle h;
	ToDoDBRecord *p;
	UInt32 uniqueID;
	UInt16 todoIndex;
	UInt16 attr;
	Char name[dmCategoryLength];

	DBGMSG((DBB, "TaskGetToDoStatus"));

	uniqueID = TaskGetToDoUniqueID(dbP, index);

	if (!uniqueID)
		return pgError;

	// try to open the ToDo database
	if (ToDoGetDatabase(&ToDoDB, dmModeReadWrite))
	{
		FrmCustomAlert(AltEmpty, "Can't open the todo database", " ", " ");
		return pgError;
	}

	if (DmFindRecordByID(ToDoDB, uniqueID, &todoIndex))
	{
		DmCloseDatabase(ToDoDB);
		return pgError;
	}

	DmRecordInfo(ToDoDB, todoIndex, &attr, NULL, NULL);
	if (attr & dmRecAttrDelete)
	{
		DmCloseDatabase(ToDoDB);
		TaskRemoveHasToDo(dbP, index);
		return pgError;
	}

// from ToDo files
//typedef struct {
	//DateType dueDate;
	//UInt8 priority;		// high bit is complete flag
	//char description;
//} ToDoDBRecord;
//#define completeFlag 	0x80
//#define priorityOnly 	~completeFlag

	h = DmQueryRecord(ToDoDB, todoIndex);
	p = MemHandleLock(h);

	if (value)
		*value = p->priority & completeFlag ? ACTION_DONE : ACTION_NOT_DONE;
	
	if (priority)
		*priority = p->priority & priorityOnly;

	if (dueDate)
	{
		*dueDate = p->dueDate;
		if (*((UInt16 *) dueDate) == toDoNoDueDate)
		{
			dueDate->month = 0;
			dueDate->year = 0;
			dueDate->day = 0;
		}
	}

	if (category)
	{
		*category = attr & dmRecAttrCategoryMask;
		if (*category == dmUnfiledCategory)
		{
			// ToDo category is Unfiled - keep Task category
			*category = TaskGetCategory(dbP, index);
		}
		else
		{
			CategoryGetName(ToDoDB, *category, name);
			*category = CategoryFind(dbP, name);
			// ToDo category not in Task - keep Task category
			if (*category == dmAllCategories)
			{
				*category = TaskGetCategory(dbP, index);
			}
		}
	}

	MemHandleUnlock(h);

	DmCloseDatabase(ToDoDB);

	return pgOK;

} // pgErr TaskGetToDoStatus(DmOpenRef dbP, UInt16 index, UInt8 *value, 
//	UInt8 *priority, DateType *dueDate, UInt16 *category)



/****************************************************************************
 * Name : TaskRepairTree
 * Desc : try to repair a broken tree
 * Prm  : 
 * 			-> start : first index to correct (0 to correct all tree)
 * 			-> end   : last index to correct (max numrec - 1)
 * Out  : 
 * Auth : lb, 30.08.2000
 ***************************************************************************/
pgErr TaskRepairTree(DmOpenRef dbP, UInt16 start, UInt16 stop)
{
	UInt16 index = 1, numRec = PgNumRecords(dbP);
	UInt8 level, nextLevel;
	UInt16 t;

	DBGMSG((DBB, "TaskRepairTree"));

	if (numRec < 2)
		return pgError;

	// this is true in all cases
	TaskSetHasPrev(dbP, 1, false);
	TaskSetHasNext(dbP, numRec - 1, false);
	TaskSetHasChild(dbP, numRec - 1, false);

	DBGMSG((DBB, "Repairing asked [%hu, %hu]", start, stop));

	if (start == 0)
	{
		// ensure a good start
		TaskSetLevel(dbP, 1, 1);

		// correct the levels + hasChild status
		for (index = 1; index < numRec - 1; index++)
		{
			level = TaskGetLevel(dbP, index);
			nextLevel = TaskGetLevel(dbP, index + 1);

			if (nextLevel == 0)
			{
				nextLevel = 1;
			}
			if (nextLevel > level)
			{
				TaskSetLevel(dbP, index + 1, level + 1);
				TaskSetHasChild(dbP, index, true);
			}
			else
			{
				TaskSetHasChild(dbP, index, false);
			}
		}
		start = 1;
		stop = numRec - 1;
	}
	else
	{
		stop += 1; // we check on further to update the next
		if (stop > numRec - 1)
			stop = numRec - 1;
		// for the previous one, we need to find it first
		// that's done at the end
	}

	DBGMSG((DBB, "Performing repair in [%hu, %hu]", start, stop));

	// correct the hasNext + hasPrev status
	for (index = start; index < stop; index++)
	{
		level = TaskGetLevel(dbP, index);
		nextLevel = TaskGetLevel(dbP, index + 1);

		if (nextLevel == level)
		{
			TaskSetHasNext(dbP, index, true);
			TaskSetHasPrev(dbP, index + 1, true);
		}
		else if (nextLevel < level)
		{
			TaskSetHasNext(dbP, index, false);
		}
		else
		{
			TaskSetHasPrev(dbP, index + 1, false);
			// need to search if there's a next below
			t = index + 1;
			while((t < numRec - 1) && TaskGetLevel(dbP, t) > level)
			{
				t++;
			}
			if (TaskGetLevel(dbP, t) == level)
			{
				TaskSetHasNext(dbP, index, true);
				TaskSetHasPrev(dbP, t, true);
			}
			else
			{
				TaskSetHasNext(dbP, index, false);
			}
		}
	}

	DBGMSG((DBB, "First loop passed"));

	if (start > 1)
	{
		UInt16 prev;
		TaskSetHasPrev(dbP, start, true); // to force the search
		prev = TaskGetPrevIndex(dbP, start);
		DBGMSG((DBB, "First passed has prev %hu", prev));
		if (prev != dmMaxRecordIndex)
		{
			TaskSetHasNext(dbP, prev, true);
		}
		else
		{
			TaskSetHasPrev(dbP, start, false);
		}
	}

	return pgOK;
} // pgErr TaskRepairTree(DmOpenRef dbP, UInt16 start, UInt16 stop)



/****************************************************************************
 * Name : TaskToClipboard
 * Desc : copy the given task (with children) to the clipboard
 * Prm  : 
 * 			-> database pointer
 * 			-> index of the task to copy
 * 			-> true => copy parent too
 * Out  : 
 * Auth : lb, 04.09.2000
 ***************************************************************************/
pgErr TaskToClipboard(DmOpenRef dbP, UInt16 index, Boolean withParent)
{
	UInt16 clipIndex = 0, i = index, numRec = PgNumRecords(dbP);
	UInt8 refLevel, level = 0;
	MemHandle hSrc, hClip;
	TaskRecordType *pSrc, *pClip;
	Err err;

	DBGMSG((DBB, "TaskToClipboard"));

	err = NewClipboard();
	if (err)
	{
		DEBUG1("Cannot create the clipboard, copy canceled.");
		return pgError;
	}

	refLevel = TaskGetLevel(dbP, index);

	// if we don't want the parent, begin at next one
	if (!withParent)
	{
		i++;
	}

	// copy the task and all its children
	while ((i < numRec && (level = TaskGetLevel(dbP, i)) > refLevel) || 
		i == index)
	{
		clipIndex = dmMaxRecordIndex; // add to the end of the clip
		hSrc = DmQueryRecord(dbP, i);
		pSrc = MemHandleLock(hSrc);

		hClip = DmNewRecord(gClip, &clipIndex, MemPtrSize(pSrc));
		MemHandleUnlock(hSrc);
		pSrc = MemHandleLock(hSrc);
		pClip = MemHandleLock(hClip);

		DmWrite(pClip, 0, pSrc, MemPtrSize(pSrc));

		MemHandleUnlock(hClip);
		MemHandleUnlock(hSrc);

		DmReleaseRecord(gClip, clipIndex, true);

		// adapt the level
		TaskSetLevel(gClip, clipIndex, level - refLevel + 1);
		
		// next record
		i++;
	}

	CloseClipboard();
	return pgOK;
} // pgErr TaskToClipboard(DmOpenRef dbP, UInt16 index, Boolean withParent)



/****************************************************************************
 * Name : TaskFromClipboard
 * Desc : copy the clipboard as a brother of the given task
 * Prm  : 
 * 			-> database pointer
 * 			-> index of the brother task
 * 			-> true to paste as children of the given task
 * Out  :   <- index of the newly pasted subtree, or -1 if error.
 * Auth : lb, 04.09.2000
 * Mod  : hcc, 07.05.2001
 *          Return the index of the newly pasted subtree
 * Mod  : lb, 2001-08-23
 * 			Correct bug : crash on paste as last child of a branch
 * 				don't return dmMaxRecordIndex
 * 			update progress after paste
 ***************************************************************************/
UInt16 TaskFromClipboard(DmOpenRef dbP, UInt16 index, Boolean asChildren)
{
	UInt16 numRec, i = 1, insertIndex = index, newlyPastedIndex=0;
	TaskRecordType *pClip, *pDest;
	MemHandle hClip, hDest;
	UInt8 refLevel = 0;
	Err err;

	DBGMSG((DBB, "TaskFromClipboard"));

	err = OpenClipboard();
	if (err)
	{
		DEBUG1("Cannot open the clipboard, paste canceled.");
		return pgError;
	}

	if ((numRec = DmNumRecords(gClip)) < 2)
	{
		CloseClipboard();
		DEBUG1("Clipboard empty");
		return pgOK;
	}

	// insert after selected one, as a brother
	index = TaskGetNextRelativeIndex(dbP, index);
	DBGMSG((DBB, "Pasting at %hu", index));
	if (!index)
		index = PgNumRecords(dbP);
	DBGMSG((DBB, "Modified to paste at %hu", index));

	while (i < numRec)
	{
		hClip = DmQueryRecord(gClip, i);
		pClip = MemHandleLock(hClip);
		hDest = DmNewRecord(dbP, &index, MemPtrSize(pClip));
		if (i == 1)
			newlyPastedIndex = index; // take the first insert index

		// reget handle (chunk could have moved, I think...)
		MemHandleUnlock(hClip);
		pClip = MemHandleLock(hClip);

		pDest = MemHandleLock(hDest);
		DmWrite(pDest, 0, pClip, MemPtrSize(pClip));
		MemHandleUnlock(hDest);
		DmReleaseRecord(dbP, index, true);
		MemHandleUnlock(hClip);

		if (i == 1)
		{
			if (asChildren)
			{
				TaskSetLevel(dbP, index, 
					(refLevel = TaskGetLevel(dbP, insertIndex) + 1));
				refLevel -= TaskGetLevel(gClip, 1);
				TaskSetHasChild(dbP, insertIndex, true);
			}
			else
			{
				TaskSetLevel(dbP, index, 
					(refLevel = TaskGetLevel(dbP, insertIndex)));
				refLevel -= TaskGetLevel(gClip, 1);
			}
		}
		else
		{
			TaskSetLevel(dbP, index, pClip->attr.bits.level + refLevel);
		}

		i++;
		index++;
	}
	DBGMSG((DBB, "Copy ok, go to repair"));
	TaskRepairTree(dbP, newlyPastedIndex, index - 1);
	TaskCalcCompleted(dbP, newlyPastedIndex);

	CloseClipboard();
	return(newlyPastedIndex);
} // pgErr TaskFromClipboard(DmOpenRef dbP, UInt16 index, Boolean asChildren)

/****************************************************************************
 * Name : QueryDeleteTask
 * Desc : Ask user if they want to delete task.
 * Parm : 
 * 		DmOpenRef dbP		-> Database pointer
 * 		UInt16 taskToDelete	-> Task to be deleted
 * Out  : 	Boolean			-> True if we should do the delete
 * Auth : rprice, 08.13.2003
 ***************************************************************************/

Boolean QueryDeleteTask(DmOpenRef dbP, UInt16 taskToDelete)
{
	Boolean deleteOK = true;

	if (!taskToDelete)
		return false;

	if (TaskGetHasChild(dbP, taskToDelete))
	{
		if (ConfirmCustom(StrActionDelete, StrTaskAndChildren) != 0)
			deleteOK=false;
	}
	else
	{
		if (gSavedPrefs.deleteWarn)
		{
			// get the description for the warning
			char* desc = TaskGetDescription(dbP, taskToDelete);
			if (FrmCustomAlert(AltConfirmDelete, desc, " ", " ") != 0)
				deleteOK = false;
			MemPtrFree(desc);
		}
	}

	return deleteOK;
}

/****************************************************************************
 * Name : TaskDeleteTree
 * Desc : Delete task (and subtasks) from database
 * Parm : 
 * 		DmOpenRef dbP		-> Database pointer
 * 		UInt16 taskToDelete	-> Task to be deleted
 * Out  : 	pgErr			-> error return
 * Auth : rprice, 08.13.2003
 ***************************************************************************/

void TaskDeleteTree(DmOpenRef dbP, UInt16 taskToDelete)
{
	UInt16 father;

	if (!taskToDelete)
		return;

	// remember the father
	father = TaskGetFatherIndex(dbP, taskToDelete);

	// remove the task
	TaskRemove(dbP, taskToDelete);
	// if father isn't a father anymore
	if (father != 0 && !TaskGetHasChild(dbP, father))
	{
		// set its precentage to 0%
		ItemType type = TaskGetType(dbP, father);

		if (type == progressType)
		{
			// was a progress, stay a progress
			TaskSetCompleted(dbP, father, 0, true);
		}
		else if (type == actionType)
		{
			// was an ACTION
			TaskSetCompleted(dbP, father, ACTION_NOT_DONE, true);
		}
		// don't touch other types
		//
		// find his father
		father = TaskGetFatherIndex(dbP, father);
	}
	// if it exists, calc its percentage
	if (father != 0)
		TaskCalcCompleted(dbP, father);
}
